@@ -14,7 +14,9 @@
  *
  * Modifications and updates by: Jesse Smith
  * Date: May 4, 2011
- * Version 1.2
+ * Version 1.2 
+ * Date: Jam 29, 2013
+ * Version 1.2 and newer
  *
  */
 
@@ -36,6 +38,20 @@
 #include <string.h>
 #include <limits.h>    // for compatibility
 
+#ifdef __APPLE__
+#include <mach/clock.h>
+#include <mach/mach.h>
+#endif
+
+
+#ifdef FREEBSD
+#include <fcntl.h>
+#include <kvm.h>
+#include <paths.h>
+#include <sys/param.h>
+#include <sys/sysctl.h>
+#include <sys/user.h>
+#endif
 
 
 //kernel time resolution (inverse of one jiffy interval) in Hertz
@@ -61,11 +77,11 @@
 #endif
 
 #ifndef VERSION
-#define VERSION 1.7
+#define VERSION 1.9
 #endif
 
 //pid of the controlled process
-pid_t pid=0;
+pid_t pid = 0;
 pid_t my_pid;     // this process's PID
 
 //executable file name
@@ -81,7 +97,9 @@ int nice_lim;
 int NCPU;
 
 //reverse byte search
-void *memrchr(const void *s, int c, size_t n);
+// void *memrchr(const void *s, int c, size_t n);
+
+
 
 //return ta-tb in microseconds (no overflow checks!)
 inline long timediff(const struct timespec *ta,const struct timespec *tb) {
@@ -294,7 +312,36 @@ void quit(int sig) {
 	exit(0);
 }
 
+
+#ifdef FREEBSD
 //get jiffies count from /proc filesystem
+int getjiffies(int pid)
+{
+   kvm_t *my_kernel = NULL;
+   struct kinfo_proc *process_data = NULL;
+   int processes;
+   int my_jiffies = -1;
+
+   my_kernel = kvm_open(0, 0, 0, O_RDONLY, "kvm_open");
+   if (! my_kernel)
+   {
+      printf("Error opening kernel vm. You should be running as root.\n");
+      return -1;
+   }
+
+   process_data = kvm_getprocs(my_kernel, KERN_PROC_PID, pid, &processes);
+   if ( (process_data) && (processes >= 1) )
+       my_jiffies = process_data->ki_runtime;
+   
+   kvm_close(my_kernel);
+   if (my_jiffies >= 0)
+     my_jiffies /= 1000;
+   return my_jiffies;
+}
+
+#endif
+
+#ifdef LINUX
 int getjiffies(int pid) {
 	static char stat[20];
 	static char buffer[1024];
@@ -321,6 +368,8 @@ int getjiffies(int pid) {
         // could not read info
         return -1;
 }
+#endif
+
 
 //process instant photo
 struct process_screenshot {
@@ -361,9 +410,22 @@ int compute_cpu_usage(int pid,int last_working_quantum,struct cpu_usage *pusage)
 	int j=getjiffies(pid);
 	if (j>=0) ps[front].jiffies=j;
 	else return -1;	//error: pid does not exist
+
+        #ifdef __APPLE__
+        // OS X does not have clock_gettime, use clock_get_time
+        clock_serv_t cclock;
+        mach_timespec_t mts;
+        host_get_clock_service(mach_host_self(), CALENDAR_CLOCK, &cclock);
+        clock_get_time(cclock, &mts);
+        mach_port_deallocate(mach_task_self(), cclock);
+        ps[front].when.tv_sec = mts.tv_sec;
+        ps[front].when.tv_nsec = mts.tv_nsec;
+
+        #else
+        // Linux and BSD can use real time
 	clock_gettime(CLOCK_REALTIME,&(ps[front].when));
 	ps[front].cputime=last_working_quantum;
-
+        #endif
 	//buffer actual size is: (front-tail+MEM_ORDER)%MEM_ORDER+1
 	int size=(front-tail+MEM_ORDER)%MEM_ORDER+1;
 
@@ -423,7 +485,7 @@ void increase_priority()
 
 void print_usage(FILE *stream,int exit_code) {
         fprintf(stream, "CPUlimit version %1.1f\n", VERSION);
-	fprintf(stream, "Usage: %s TARGET [OPTIONS...]\n",program_name);
+	fprintf(stream, "Usage: %s TARGET [OPTIONS...] [PROGRAM]\n",program_name);
 	fprintf(stream, "   TARGET must be exactly one of these:\n");
 	fprintf(stream, "      -p, --pid=N        pid of the process\n");
 	fprintf(stream, "      -e, --exe=FILE     name of the executable program file\n");
@@ -437,6 +499,11 @@ void print_usage(FILE *stream,int exit_code) {
 	fprintf(stream, "      -l, --limit=N      percentage of cpu allowed from 1 up.\n");
         fprintf(stream, "                         Usually 1 - %d00, but can be higher\n", NCPU);
         fprintf(stream, "                         on multi-core CPUs (mandatory)\n");
+        fprintf(stream, "      -k, --kill         kill processes going over their limit\n");
+        fprintf(stream, "                         instead of just throttling them.\n");
+        fprintf(stream, "      -r, --restore      Restore processes after they have\n");
+        fprintf(stream, "                         been killed. Works with the -k flag.\n");
+
 	fprintf(stream, "      -v, --verbose      show control statistics\n");
 	fprintf(stream, "      -z, --lazy         exit if there is no suitable target process,\n");
         fprintf(stream, "                         or if it dies\n");
@@ -462,13 +529,14 @@ int get_ncpu()
 int main(int argc, char **argv) {
 
 	//get program name
-	char *p=(char*)memrchr(argv[0],(unsigned int)'/',strlen(argv[0]));
-	program_name = p==NULL?argv[0]:(p+1);
+	// char *p=(char*)memrchr(argv[0],(unsigned int)'/',strlen(argv[0]));
+	// program_name = p==NULL?argv[0]:(p+1);
+        program_name = argv[0];
         int run_in_background = FALSE;
 	//parse arguments
 	int next_option;
 	/* A string listing valid short options letters. */
-	const char* short_options="p:e:P:l:c:bvzh";
+	const char* short_options="p:e:P:l:c:bkrvzh";
 	/* An array describing valid long options. */
 	const struct option long_options[] = {
 		{ "pid", required_argument, NULL, 'p' },
@@ -490,6 +558,8 @@ int main(int argc, char **argv) {
 	int process_ok = FALSE;
 	int limit_ok = FALSE;
         int last_known_argument = 0;
+        int kill_process = FALSE;   // kill process instead of stopping it
+        int restore_process = FALSE;  // restore killed process
         // struct rlimit maxlimit;
 
         NCPU = get_ncpu();
@@ -526,6 +596,15 @@ int main(int argc, char **argv) {
                                 NCPU = atoi(optarg);
                                 last_known_argument += 2;
                                 break;
+                        case 'k':
+                                kill_process = TRUE;
+                                last_known_argument++;
+                                break;
+                        case 'r':
+                                restore_process = TRUE;
+                                last_known_argument++;
+                                break;
+
 			case 'v':
 				verbose = TRUE;
                                 last_known_argument++;
@@ -578,6 +657,11 @@ int main(int argc, char **argv) {
            else     // parent who will now fork the throttler
            {
               pid_t limit_pid;
+              // if we are planning to kill a process, give it
+              // a running head start to avoid death at start-up
+              if (kill_process)
+                 sleep(5);
+     
               limit_pid = fork();
               if (limit_pid == 0)   // child
               {
@@ -714,9 +798,9 @@ gradual approach, rather than "all or nothing".
 wait_for_process:
 
 	//look for the target process..or wait for it
-	if (exe!=NULL)
+	if (exe != NULL)
 		pid=getpidof(exe);
-	else if (path!=NULL)
+	else if (path != NULL)
 		pid=getpidof(path);
 	else 
 		waitforpid(pid);
@@ -773,9 +857,15 @@ wait_for_process:
 
 		if (verbose && i%10==0 && i>0) {
 			printf("%0.2f%%\t%6ld us\t%6ld us\t%0.2f%%\n",pcpu*100,twork.tv_nsec/1000,tsleep.tv_nsec/1000,workingrate*100);
+                        if (i%200 == 0)
+                           print_caption();
 		}
 
-		if (limit<1 && limit>0) {
+		// if (limit<1 && limit>0) {
+                // printf("Comparing %f to %f\n", pcpu, limit);
+                if (pcpu < limit)
+                {
+                        // printf("Continue\n");
 			//resume process
 			if (kill(pid,SIGCONT)!=0) {
 				fprintf(stderr,"Process %d dead!\n",pid);
@@ -785,21 +875,91 @@ wait_for_process:
 			}
 		}
 
+                #ifdef __APPLE_
+                // OS X does not have clock_gettime, use clock_get_time
+                clock_serv_t cclock;
+                mach_timespec_t mts;
+                host_get_clock_service(mach_host_self(), CALENDAR_CLOCK, &cclock);
+                clock_get_time(cclock, &mts);
+                mach_port_deallocate(mach_task_self(), cclock);
+                startwork.tv_sec = mts.tv_sec;
+                startwork.tv_nsec = mts.tv_nsec;
+
+                #else
 		clock_gettime(CLOCK_REALTIME,&startwork);
-		nanosleep(&twork,NULL);		//now process is working	
+                #endif
+
+		nanosleep(&twork,NULL);		//now process is working
+                #ifdef __APPLE__
+                // OS X does not have clock_gettime, use clock_get_time
+                // clock_serv_t cclock;
+                // mach_timespec_t mts;
+                host_get_clock_service(mach_host_self(), CALENDAR_CLOCK, &cclock);
+                clock_get_time(cclock, &mts);
+                mach_port_deallocate(mach_task_self(), cclock);
+                endwork.tv_sec = mts.tv_sec;
+                endwork.tv_nsec = mts.tv_nsec;
+
+                #else
 		clock_gettime(CLOCK_REALTIME,&endwork);
+                #endif
 		workingtime=timediff(&endwork,&startwork);
 
-		if (limit<1) {
+		// if (limit<1) {
+                // printf("Checking %f vs %f\n", pcpu, limit);
+                if (pcpu > limit)
+                {
+                     // When over our limit we may run into
+                     // situations where we want to kill
+                     // the offending process, then restart it
+                     if (kill_process)
+                     {
+                         kill(pid, SIGKILL);
+                         fprintf(stderr, "Process %d killed.\n", pid);
+                         if ( (lazy) && (! restore_process) ) 
+                              exit(2);
+                         // restart killed process
+                         if (restore_process)
+                         {
+                             pid_t new_process;
+                             new_process = fork();
+                             if (new_process == -1)
+                             {
+                              fprintf(stderr, "Failed to restore killed process.\n");
+                             }
+                             else if (new_process == 0)
+                             {
+                                // child which becomes new process
+                                if (verbose)
+                                   printf("Relaunching %s\n",
+                                          argv[last_known_argument]);
+                                execvp(argv[last_known_argument],
+                                       &(argv[last_known_argument]) ); 
+                             }
+                             else // parent
+                             {
+                                // we need to track new process
+                                pid = new_process;
+                                // avoid killing child process
+                                sleep(5);
+                             }
+                         }
+                     }
+                     // do not kll process, just throttle it
+                     else
+                     {
+
+                        // printf("Stop\n");
 			//stop process, it has worked enough
 			if (kill(pid,SIGSTOP)!=0) {
-				fprintf(stderr,"Process %d dead!\n",pid);
+				fprintf(stderr,"Process %d dead!\n", pid);
 				if (lazy) exit(2);
 				//wait until our process appears
 				goto wait_for_process;
 			}
 			nanosleep(&tsleep,NULL);	//now process is sleeping
-		}
+                      }   // end of throttle process
+		}         // end of process using too much CPU
 		i++;
 	}
 
